import { GraphType, LinkType, NodeType } from '../Types';
import { tailwindColors } from 'config';
import { ReactEventHandler, useEffect, useMemo, useRef, useState } from 'react';
import { Application, Circle, Container, FederatedPointerEvent, Graphics, IPointData } from 'pixi.js';
import { binaryColor, nodeColor as nodeColor } from './utils';
import { select, zoom as d3zoom, drag as d3drag } from 'd3';
import * as force from './NLForce';
import { Viewport } from 'pixi-viewport';
import { GraphQueryResult, GraphQueryResultFromBackendPayload } from '@graphpolaris/shared/lib/data-access/store/graphQueryResultSlice';
import { useAppDispatch, useML } from '@graphpolaris/shared/lib/data-access';
import { ML, setShortestPathSource, setShortestPathTarget } from '@graphpolaris/shared/lib/data-access/store/mlSlice';
import { parseQueryResult } from './query2NL';

type Props = {
  onClick: (node: NodeType) => void;
  onHover: (data: { node: NodeType; pos: IPointData } | undefined) => void;
  highlightNodes: NodeType[];
  currentShortestPathEdges?: LinkType[];
  highlightedLinks?: LinkType[];
  graph?: GraphType;
};

const app = new Application({ background: 0xffffff, antialias: true, autoDensity: true, eventMode: 'auto' });
const nodes = new Container();
const links = new Container();

//////////////////
// MAIN COMPONENT
//////////////////

export const NLPixi = (props: Props) => {
  const nodeMap = useRef(new Map<string, Graphics>());
  const linkMap = useRef(new Map<string, Graphics>());
  const viewport = useRef<Viewport>();
  const ref = useRef<HTMLDivElement>(null);
  const isSetup = useRef(false);
  const ml = useML();
  const dragging = useRef<{ node: NodeType; gfx: Graphics } | null>(null);
  const onlyClicked = useRef(false);
  const dispatch = useAppDispatch();

  // useEffect(() => {
  //   app.renderer.resize(props.windowSize.width, props.windowSize.height);
  //   app.render();
  // }, [props.windowSize]);

  function resize() {
    const width = ref?.current?.clientWidth || 1000;
    const height = ref?.current?.clientHeight || 1000;
    app.renderer.resize(width, height);
    if (viewport.current) {
      viewport.current.screenWidth = width;
      viewport.current.worldWidth = width;
      viewport.current.worldHeight = height;
      viewport.current.screenHeight = height;
    }
    app.render();
  }

  useEffect(() => {
    if (!ref.current) return;
    const resizeObserver = new ResizeObserver(() => {
      resize();
    });
    resizeObserver.observe(ref.current);
    return () => resizeObserver.disconnect(); // clean up
  }, []);

  useEffect(() => {
    if (ref.current && ref.current.children.length === 0) {
      ref.current.appendChild(app.view as HTMLCanvasElement);
      resize();
    }
  }, [ref]);

  function onDragStart(event: FederatedPointerEvent, node: NodeType, gfx: Graphics) {
    // store a reference to the data
    // the reason for this is because of multitouch
    // we want to track the movement of this particular touch
    event.stopPropagation();
    console.log('dragstart', node);
    onHover(event, undefined);

    if (viewport.current) viewport.current.pause = true;
    dragging.current = { node, gfx };
    onlyClicked.current = true;
  }

  function onHover(event: FederatedPointerEvent, node: NodeType | undefined) {
    event.stopPropagation();
    if (viewport?.current && !viewport?.current?.pause && node) {
      const rect = ref.current?.getBoundingClientRect();
      const x = (rect?.x || 0) + (node.x || 0);
      const y = (rect?.y || 0) + (node.y || 0);
      props.onHover({ node, pos: viewport.current.toScreen(x, y) });
    } else {
      props.onHover(undefined);
    }
  }

  function onDragEnd(event: FederatedPointerEvent) {
    if (dragging.current) {
      event.stopPropagation();
      dragging.current.node.fx = null;
      dragging.current.node.fy = null;
      if (viewport.current) viewport.current.pause = false;
      if (onlyClicked.current) {
        onlyClicked.current = false;
        props.onClick(dragging.current.node);
      }
      onHover(event, dragging.current.node);
      dragging.current = null;
    }
  }

  function onDragMove(event: FederatedPointerEvent) {
    if (dragging.current) {
      onlyClicked.current = false;
      event.stopPropagation();

      if (!dragging.current.node.fx) dragging.current.node.fx = dragging.current.node.x || 0;
      if (!dragging.current.node.fy) dragging.current.node.fy = dragging.current.node.y || 0;
      dragging.current.node.fx += event.movementX / (viewport.current?.scaled || 1);
      dragging.current.node.fy += event.movementY / (viewport.current?.scaled || 1);
      force.simulation.alpha(0.1).restart();
    }
  }

  const updateNode = (node: NodeType) => {
    const gfx = nodeMap.current.get(node.id);
    if (!gfx) return;

    const lineColor = node.isShortestPathSource
      ? tailwindColors.entity[950]
      : node.isShortestPathTarget
      ? tailwindColors.relation[950]
      : node.selected
      ? tailwindColors.entity[400]
      : '#000000';
    const lineWidth = node.selected ? 3 : 1.5;
    gfx.lineStyle(lineWidth, binaryColor(lineColor));

    if (node?.cluster) {
      gfx.beginFill(node.cluster >= 0 ? nodeColor(node.cluster) : 0x000000);
    } else gfx.beginFill(nodeColor(node.type));

    gfx.drawCircle(0, 0, Math.max(node.radius || 5, 5));
    gfx.endFill();

    gfx.position.set(node.x, node.y);

    gfx.off('mouseover');
    gfx.off('mousedown');
    gfx.on('mouseover', (e) => onHover(e, node));
    gfx.on('mouseout', (e) => onHover(e, undefined));
    gfx.on('mousedown', (e) => onDragStart(e, node, gfx));

    // if (!item.position) {
    //   item.position = new Point(node.x, node.y);
    // } else {
    //   item.position.set(node.x, node.y);
    // }
    // Update attributes position if they exist
    // if (node.gfxAttributes) {
    //   const x = node.x - node.gfxAttributes.width / 2;
    //   const y = node.y - node.gfxAttributes.height - 20;
    //   if (!node.gfxAttributes?.position) node.gfxAttributes.position = new Point(x, y);
    //   else {
    //     node.gfxAttributes.position.set(x, y);
    //   }
    // }
  };

  const createNode = (node: NodeType, selected?: boolean) => {
    // check if node is already drawn, and if so, delete it
    if (node && node?.id && nodeMap.current?.has(node.id)) {
      nodeMap.current.delete(node.id);
    }
    // Do not draw node if it has no position
    if (node.x === undefined || node.y === undefined) return;
    const gfx = new Graphics();
    nodeMap.current.set(node.id, gfx);
    nodes.addChild(gfx);
    node.selected = selected;

    updateNode(node);
    gfx.name = 'node_' + node.id;
    gfx.eventMode = 'dynamic';

    return gfx;
  };

  // /** UpdateRadius works just like UpdateColors, but also applies radius*/
  // const UpdateRadius = (graph: GraphType, radius: number) => {
  //   // update for each node in graph
  //   graph.nodes.forEach((node: NodeType) => {
  //     createNode(node);
  //   });
  // };

  const updateLink = (link: LinkType, gfx: Graphics) => {
    if (!props.graph) return;

    const _source = link.source;
    const _target = link.target;

    if (!_source || !_target) {
      console.log('source or target not found', _source, _target);
      return;
    }

    let sourceId = '';
    let targetId = '';
    let source: NodeType | undefined;
    let target: NodeType | undefined;

    if (typeof _source === 'string') {
      sourceId = link.source as string;
      targetId = link.target as string;
      source = nodeMap.current.get(sourceId) as NodeType | undefined;
      target = nodeMap.current.get(targetId) as NodeType | undefined;
    } else {
      source = link.source as NodeType;
      target = link.target as NodeType;
      sourceId = source.id;
      targetId = target.id;
    }
    if (!source || !target) {
      console.error('source or target not found', source, target, sourceId, targetId);
      return;
    }

    if (gfx) {
      let color = link.color || 0x000000;
      let style = 0.2;
      let alpha = link.alpha || 1;
      if (link.mlEdge) {
        color = 0x0000ff;
        if (link.value > ml.communityDetection.jaccard_threshold) {
          style = link.value * 1.8;
        } else {
          style = 0;
          alpha = 0.2;
        }
      } else if (props.highlightedLinks && props.highlightedLinks.includes(link)) {
        if (link.mlEdge && ml.communityDetection.jaccard_threshold) {
          if (link.value > ml.communityDetection.jaccard_threshold) {
            color = 0xaa00ff;
            style = link.value * 1.8;
          }
        } else {
          color = 0xff0000;
          style = 1.0;
        }
      } else if (props.currentShortestPathEdges && props.currentShortestPathEdges.includes(link)) {
        color = 0x00ff00;
        style = 3.0;
      }

      gfx.clear();
      gfx.beginFill();
      gfx
        .lineStyle(style, color, alpha)
        .moveTo(source.x || 0, source.y || 0)
        .lineTo(target.x || 0, target.y || 0);
      gfx.endFill();
    } else {
      throw Error('Link not found');
    }
  };

  const createLink = (link: LinkType) => {
    if (linkMap.current.has(link.id)) {
      linkMap.current.delete(link.id);
    }
    const gfx = new Graphics();
    gfx.name = 'link_' + link.id;
    linkMap.current.set(link.id, gfx);
    updateLink(link, gfx);
    links.addChild(gfx);
    return gfx;
  };

  useEffect(() => {
    if (props.graph && ref.current && ref.current.children.length > 0) {
      if (!isSetup.current) setup();
      else update();
      // setup();
    }
  }, [props.graph]);

  const tick = (delta: number) => {
    if (props.graph) {
      props.graph.nodes.forEach((node: NodeType) => {
        const gfx = nodeMap.current.get(node.id);
        if (!gfx || node.x === undefined || node.y === undefined) return;
        gfx.position.copyFrom(node as IPointData);
      });

      // Update forces of the links
      props.graph.links.forEach((link: any) => {
        if (linkMap.current && !!linkMap.current.has(link.id)) updateLink(link, linkMap.current.get(link.id) as Graphics);
      });
    }
  };

  const update = (forceClear = false) => {
    if (!props.graph || !ref.current) return;

    if (props.graph) {
      if (forceClear) {
        nodeMap.current.clear();
        linkMap.current.clear();
        nodes.removeChildren();
        links.removeChildren();
      }

      nodeMap.current.forEach((gfx, id) => {
        if (!props.graph?.nodes?.find((node) => node.id === id)) {
          nodes.removeChild(gfx);
          gfx.destroy();
          nodeMap.current.delete(id);
        }
      });

      linkMap.current.forEach((gfx, id) => {
        if (!props.graph?.links?.find((link) => link.id === id)) {
          links.removeChild(gfx);
          gfx.destroy();
          linkMap.current.delete(id);
        }
      });

      props.graph.nodes.forEach((node: NodeType) => {
        if (!forceClear && nodeMap.current.has(node.id)) {
          const old = nodeMap.current.get(node.id);
          node.x = old?.x || node.x;
          node.y = old?.y || node.y;
          updateNode(node);
        } else {
          createNode(node);
        }
      });
      props.graph.links.forEach((link: LinkType) => {
        if (!forceClear && linkMap.current.has(link.id)) {
          const gfx = linkMap.current.get(link.id);
          if (gfx) updateLink(link, gfx);
        } else createLink(link);
      });

      // // update text colour (written after nodes so that text appears on top of nodes)
      //   nodes.forEach((node: NodeType) => {
      //   if (node.gfxAttributes !== undefined) {
      //       const selected = node.selected === true;
      //       node.gfxAttributes.destroy();
      //       createAttributes(node);
      //       if (selected) {
      //       showAttributes(node);
      //       }
      //   }
      //   });
      // console.log(nodes.children);

      force.startSimulation(props.graph, ref.current.getBoundingClientRect());
      force.simulation.on('tick', () => {});
      app.ticker.add(tick);
    }
  };

  /**
   * SetNodeGraphics is an initializing function. It attaches the nodes and links to the simulation.
   * It creates graphic objects and adds these to the PIXI containers. It also clears both of these of previous nodes and links.
   * @param graph The graph returned from the database and that is parsed into a nodelist and edgelist.
   */
  const setup = () => {
    nodes.removeChildren();
    links.removeChildren();
    app.stage.removeChildren();

    if (!props.graph) throw Error('Graph is undefined');

    const size = ref.current?.getBoundingClientRect();
    viewport.current = new Viewport({
      screenWidth: size?.width || 1000,
      screenHeight: size?.height || 1000,
      worldWidth: size?.width || 1000,
      worldHeight: size?.height || 1000,
      stopPropagation: true,
      events: app.renderer.events, // the interaction module is important for wheel to work properly when renderer.view is placed or scaled
    });
    app.stage.addChild(viewport.current);
    // activate plugins
    viewport.current.drag().pinch().wheel().animate({}).decelerate({ friction: 0.75 });

    viewport.current.addChild(links);
    viewport.current.addChild(nodes);
    // app.stage.addChild(links);
    // app.stage.addChild(nodes);
    app.stage.eventMode = 'dynamic';
    app.stage.on('mouseup', onDragEnd);
    app.stage.on('pointerup', onDragEnd);
    app.stage.on('mousemove', onDragMove);
    app.stage.on('mouseup', onDragEnd);

    app.ticker.add(tick);

    nodeMap.current.clear();
    linkMap.current.clear();
    update();
    isSetup.current = true;
  };

  return <div className="h-full w-full overflow-hidden" ref={ref}></div>;
};
